"""Inspired by event.py from https://github.com/jeffthibault/python-nostr.git."""
import binascii
import datetime
import json
import time
from dataclasses import dataclass
from enum import IntEnum
from hashlib import sha256
from typing import List, Optional

from .key import PrivateKey, PublicKey
from .message_type import ClientMessageType
from .utils import bech32_encode


class EventKind(IntEnum):
    SET_METADATA = 0
    TEXT_NOTE = 1
    RECOMMEND_RELAY = 2
    CONTACTS = 3
    ENCRYPTED_DIRECT_MESSAGE = 4
    DELETE = 5
    REACTION = 7
    BADGE_AWARD = 8
    CHANNEL_CREATE = 40
    CHANNEL_META = 41
    CHANNEL_MESSAGE = 42
    CHANNEL_HIDE = 43
    CHANNEL_MUTE = 44
    REPORT = 1984
    ZAP_REQUEST = 9734
    ZAPPER = 9735
    RELAY_LIST_METADATA = 10002
    PROFILE_BADGES = 30008
    BADGE_DEFINITION = 30009
    LONG_FORM_CONTENT = 30023


@dataclass
class Event:
    """Event class.

    :param content: content string
    :param pukey: public key in hex form
    :param created_at: event creation date
    :param kind: event kind
    :param tags: list of list of strings
    :param id: event id, will be computed
    :param sig: signature, will be created after signing with a private key
    """

    content: Optional[str] = None
    pubkey: Optional[str] = None
    created_at: Optional[int] = None
    kind: Optional[int] = EventKind.TEXT_NOTE
    tags: List[List[str]] = None
    id: Optional[str] = None
    sig: Optional[str] = None

    def __post_init__(self):
        if self.content is not None and not isinstance(self.content, str):
            # DMs initialize content to None but all other kinds should pass in a str
            raise TypeError("Argument 'content' must be of type str")
        if self.created_at is None:
            self.created_at = int(time.time())

        if self.tags is None:
            self.tags = []

        if self.id is None:
            self.compute_id()

    def serialize(self) -> bytes:
        data = [0, self.pubkey, self.created_at, self.kind, self.tags, self.content]
        data_str = json.dumps(data, separators=(',', ':'), ensure_ascii=False)
        return data_str.encode()

    def compute_id(self):
        self.id = sha256(self.serialize()).hexdigest()

    @classmethod
    def from_dict(cls, msg: dict) -> 'Event':
        # "id" is ignore, as it will be computed from the contents
        return Event(
            content=msg['content'],
            pubkey=msg['pubkey'],
            created_at=msg['created_at'],
            kind=msg['kind'],
            tags=msg['tags'],
            sig=msg['sig'],
        )

    def add_pubkey_ref(self, pubkey: str):
        """Adds a reference to a pubkey as a 'p' tag."""
        self.tags.append(['p', pubkey])
        self.compute_id()

    def has_pubkey_ref(self, pubkey: str):
        for tag in self.tags:
            if not tag and len(tag) < 2:
                continue
            if tag[0] == 'p' and tag[1] == pubkey:
                return True
        return False

    def add_event_ref(self, event_id: str):
        """Adds a reference to an event_id as an 'e' tag."""
        self.tags.append(['e', event_id])
        self.compute_id()

    def has_event_ref(self, event_id: str):
        """Check if a e tag to the given event_id exists."""
        for tag in self.tags:
            if not tag and len(tag) < 2:
                continue
            if tag[0] == 'e' and tag[1] == event_id:
                return True
        return False

    def get_tag_dict(self):
        """Returns all tags as dict."""
        ret = {}
        tag_types = self.get_tag_types()
        for t in tag_types:
            ret[t] = self.get_tag_list(tag_type=t)
        return ret

    def get_tag_list(self, tag_type: str = 'e'):
        """Returns all tags of given type as list."""
        ret = []
        for tag in self.tags:
            if not tag:
                continue
            if tag[0] == tag_type:
                ret.append(tag[1])
        return ret

    def get_tag_types(self):
        """Returns list of all included tag types."""
        ret = []
        for tag in self.tags:
            if not tag and len(tag) == 0:
                continue
            if tag[0] not in ret:
                ret.append(tag[0])
        return ret

    def get_tag_count(self, tag_type: str = 'e'):
        """Returns all tags of given type as list."""
        count = 0
        for tag in self.tags:
            if len(tag) > 0 and tag[0] == tag_type:
                count += 1
        return count

    def bech32(self, prefix: str = "note") -> str:
        """bech32-encoded entities (N-19)

        :return: note id as bech32 encoding with note prefix
        """
        self.compute_id()
        return bech32_encode(binascii.unhexlify(self.id), prefix)

    def sign(self, private_key_hex: str) -> None:
        """signs the event with the private key and stored the signature in self.sig.
        The pubkey from the event is replaced and the note id recomputed.

        :param private_key_hex: private key as hex string
        """
        if self.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE and self.content is None:
            raise Exception("Message is not yet encrypted!")
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        self.pubkey = sk.public_key.hex()
        self.compute_id()
        sig = sk.sign(bytes.fromhex(self.id))
        self.sig = sig.hex()

    def verify(self) -> bool:
        pub_key = PublicKey.from_hex(self.pubkey)
        self.compute_id()
        return pub_key.verify(bytes.fromhex(self.sig), bytes.fromhex(self.id))

    def date_time(self):
        return datetime.datetime.utcfromtimestamp(self.created_at)

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "pubkey": self.pubkey,
            "created_at": self.created_at,
            "kind": self.kind,
            "tags": self.tags,
            "content": self.content,
            "sig": self.sig,
        }

    def to_message(self) -> str:
        return json.dumps(
            [
                ClientMessageType.EVENT,
                self.to_dict(),
            ]
        )

    def __repr__(self):
        return f'Event({self.id[:10]}...{self.id[-10:]})'

    def __str__(self):
        return self.to_message()



"""Forked from https://github.com/jeffthibault/python-nostr.git."""
import json
import threading
import time
from dataclasses import dataclass
from threading import Lock

from .base_relay import RelayPolicy, RelayProxyConnectionConfig
from .event import Event
from .exception import RelayException
from .filters import FiltersList
from .message_pool import MessagePool
from .websocket_relay import WebSocketRelay


@dataclass
class WebSocketRelayManager:
    error_threshold: int = 0
    connection_monitor_interval_secs: int = 5

    def __post_init__(self):
        self.relays: dict[str, WebSocketRelay] = {}
        self.message_pool: MessagePool = MessagePool()
        self.lock: Lock = Lock()

        threading.Thread(
            target=self._relay_connection_monitor,
            name="relay-connection-monitor",
            daemon=True,
        ).start()

    def add_relay(
        self,
        url: str,
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: RelayProxyConnectionConfig = None,
    ):

        relay = WebSocketRelay(
            url, self.message_pool, policy, ssl_options, proxy_config
        )
        if self.error_threshold:
            relay.error_threshold = self.error_threshold

        with self.lock:
            self.relays[url] = relay
            relay.connect()

    def remove_relay(self, url: str):
        with self.lock:
            if url in self.relays:
                relay = self.relays.pop(url)
                relay.close()

    def _relay_connection_monitor(self):
        while True:
            with self.lock:
                for relay in self.relays.values():
                    if not relay.is_connected:
                        relay.connect(True)

            time.sleep(self.connection_monitor_interval_secs)

    def remove_closed_relays(self):
        for url, connected in self.connection_statuses.items():
            if not connected:
                # warnings.warn(f'{url} is not connected... removing relay.')
                self.remove_relay(url=url)

    def add_subscription_on_relay(self, url: str, id: str, filters: FiltersList):
        with self.lock:
            if url in self.relays:
                relay = self.relays[url]
                if not relay.policy.should_read:
                    raise RelayException(
                        f"Could not send request: {url} "
                        f"is not configured to read from"
                    )
                relay.add_subscription(id, filters)
                relay.publish(relay.request)
            else:
                raise RelayException(f"Invalid relay url: no connection to {url}")

    def add_subscription_on_all_relays(self, id: str, filters: FiltersList):
        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_read:
                    relay.add_subscription(id, filters)
                    relay.publish(relay.request)

    def close_subscription_on_relay(self, url: str, id: str):
        with self.lock:
            if url in self.relays:
                relay = self.relays[url]
                relay.close_subscription(id)
                relay.publish(json.dumps(["CLOSE", id]))
            else:
                raise RelayException(f"Invalid relay url: no connection to {url}")

    def close_subscription_on_all_relays(self, id: str):
        with self.lock:
            for relay in self.relays.values():
                relay.close_subscription(id)
                relay.publish(json.dumps(["CLOSE", id]))

    def close_all_relay_connections(self):
        with self.lock:
            for url in self.relays:
                relay = self.relays[url]
                relay.close()

    @property
    def connection_statuses(self) -> dict:
        """gets the url and connection statuses of relays
        Returns:
            dict: bool of connection statuses
        """
        statuses = [relay.is_connected for relay in self.relays.values()]
        return dict(zip(self.relays.keys(), statuses))

    def publish_message(self, message: str):
        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_write:
                    relay.publish(message)

    def publish_event(self, event: Event):
        """Verifies that the Event is publishable before submitting it to relays."""
        if event.sig is None:
            raise RelayException(f"Could not publish {event.id}: must be signed")

        if not event.verify():
            raise RelayException(
                f"Could not publish {event.id}: failed to verify signature {event.sig}"
            )

        self.publish_message(event.to_message())



import logging
from dataclasses import dataclass
from typing import Optional

from .event import Event, EventKind
from .exception import NIPValidationException
from .key import PrivateKey, PublicKey

log = logging.getLogger(__name__)


@dataclass
class EncryptedDirectMessage:
    """NIP-04 Encrypted Direct Message."""

    pubkey: Optional[str] = None
    recipient_pubkey: Optional[str] = None
    cleartext_content: Optional[str] = None
    encrypted_message: Optional[str] = None
    reference_event_id: Optional[str] = None

    @classmethod
    def from_event(cls, event: Event) -> 'EncryptedDirectMessage':
        if event.kind != EventKind.ENCRYPTED_DIRECT_MESSAGE:
            return None
        dm = EncryptedDirectMessage(encrypted_message=event.content)
        dm.pubkey = event.pubkey
        dm.event = event
        return dm

    @classmethod
    def from_npub(cls, npub: str) -> 'EncryptedDirectMessage':
        dm = EncryptedDirectMessage()
        dm.pubkey = PublicKey.from_npub(npub).hex()
        return dm

    def to_event(self) -> Event:
        e = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        if self.encrypted_message is None or "?iv=" not in self.encrypted_message:
            raise NIPValidationException("Encrypted message is missing!")
        if self.recipient_pubkey is None:
            raise NIPValidationException("recipient_pubkey is missing!")
        e.content = self.encrypted_message
        e.pubkey = self.pubkey
        e.add_pubkey_ref(self.recipient_pubkey)
        if self.reference_event_id is not None:
            e.add_event_ref(self.reference_event_id)
        e.compute_id()
        return e

    def encrypt(
        self,
        private_key_hex: str,
        cleartext_content: str = None,
        recipient_pubkey: str = None,
    ) -> None:
        if cleartext_content is not None:
            self.cleartext_content = cleartext_content
        if recipient_pubkey is not None:
            self.recipient_pubkey = recipient_pubkey
        if self.recipient_pubkey is None:
            raise Exception("recipient_pubkey must not be None")
        if self.cleartext_content is None:
            raise Exception("cleartext_content must not be None")
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        self.pubkey = sk.public_key.hex()
        self.encrypted_message = sk.encrypt_message(
            message=self.cleartext_content, public_key_hex=self.recipient_pubkey
        )

    def decrypt(
        self,
        private_key_hex: str,
        encrypted_message: str = None,
        public_key_hex: str = None,
    ) -> None:
        if encrypted_message is not None:
            self.encrypted_message = encrypted_message
        if public_key_hex is None:
            public_key_hex = self.recipient_pubkey
        if public_key_hex is None:
            raise Exception("public_key must not be None")
        if self.encrypted_message is None:
            raise Exception("encrypted_message must not be None")
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        self.cleartext_content = sk.decrypt_message(
            encoded_message=self.encrypted_message, public_key_hex=public_key_hex
        )



import unittest

from pynostr import pow
from pynostr.event import Event, EventKind
from pynostr.key import PrivateKey
from pynostr.pow import Pow, PowEvent, PowKey, PowVanityKey


class TestPow(unittest.TestCase):
    def test_count_leading_zero(self):
        self.assertEqual(pow.count_leading_zero_bits("048d"), 5)
        self.assertEqual(pow.count_leading_zero_bits("ffff"), 0)

    def test_mine_event(self):
        """Test mining an event with specific difficulty."""
        public_key = PrivateKey().public_key.hex()
        difficulty = 8
        event = Event(content='test', pubkey=public_key, kind=EventKind.TEXT_NOTE)
        p = PowEvent(difficulty)
        self.assertFalse(p.check_difficulty(event))
        event = p.mine(event)
        self.assertTrue(p.calc_difficulty(event) >= difficulty)
        self.assertTrue(p.check_difficulty(event))

    def test_check_difficulty_event(self):
        pow_e = Event.from_dict(
            {
                "id": "000006d8c378af1779d2feebc7603a125d99eca0ccf1085959b307f64e5dd"
                "358",
                "pubkey": "a48380f4cfcc1ad5378294fcac36439770f9c878dd880ffa94bb74ea54a6"
                "f243",
                "created_at": 1651794653,
                "kind": 1,
                "tags": [["nonce", "776797", "20"]],
                "content": "It's just me mining my own business",
                "sig": "284622fc0a3f4f1303455d5175f7ba962a3300d136085b9566801bc2e0699de"
                "0c7e31e44c81fb40ad9049173742e904713c3594a1da0fc5d2382a25c11aba977",
            }
        )
        p = PowEvent(20)
        self.assertTrue(p.check_difficulty(pow_e))

    def test_mine_key(self):
        """Test mining a public key with specific difficulty."""
        difficulty = 8
        p = PowKey(difficulty)
        sk = p.mine()
        self.assertTrue(pow.count_leading_zero_bits(sk.public_key.hex()) >= difficulty)

    def test_time_estimates(self):
        """Test functions to estimate POW time."""
        public_key = PrivateKey().public_key.hex()
        event = Event(content='test', pubkey=public_key, kind=EventKind.TEXT_NOTE)
        p = PowEvent(8)
        # test successful run of all estimators
        p.get_expected_time()
        p.mine(event)
        p.get_expected_time()
        p_key = PowKey(8)
        p_key.get_expected_time()
        p_vkey = PowVanityKey("t")
        p_vkey.get_expected_time()

    def test_mine_vanity_key(self):
        """Test vanity key mining."""
        pattern = '23'
        p = PowVanityKey(pattern)
        sk = p.mine()
        sk.public_key.bech32()
        self.assertTrue(sk.public_key.bech32().startswith(f'npub1{pattern}'))

        p = PowVanityKey(suffix=pattern)
        sk = p.mine()
        self.assertTrue(sk.public_key.bech32().endswith(pattern))

        # mine an invalid pattern
        pattern = '1'

        with self.assertRaisesRegex(ValueError, "not in valid list of bech32 chars"):
            p = PowVanityKey(pattern)

    def test_expected_pow_guesses(self):
        p = Pow()
        p.n_pattern = 32
        p.n_options = 2
        e1 = p.get_expected_guesses()

        p.n_pattern = 8
        p.n_options = 16
        e2 = p.get_expected_guesses()
        self.assertEqual(e1, e2)



import unittest

from pynostr.exception import NIPValidationException
from pynostr.key import PrivateKey
from pynostr.report import Report, ReportType


class TestReport(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.reported_pubkey = PrivateKey().public_key.hex()

    def test_report_type(self):
        report1 = Report(
            report_type="invalidtype", reported_pubkey=self.reported_pubkey
        )
        with self.assertRaisesRegex(NIPValidationException, "valid report type"):
            report1.to_event()
        report2 = Report(report_type=ReportType.NUDITY)
        with self.assertRaisesRegex(
            NIPValidationException, "pubkey of the user being reported"
        ):
            report2.to_event()

    def test_report_tags(self):
        report = Report(
            reported_pubkey=self.reported_pubkey,
            note_id="fakenoteid",
            report_type=ReportType.ILLEGAL,
            victim_pubkey="thevictim",
        )
        report_event = report.to_event()
        self.assertEqual(len(report_event.tags), 2)
        tag_types = [tag[0] for tag in report_event.tags]
        self.assertEqual(tag_types, ["e", "p"])



import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .event import Event, EventKind
from .exception import NIPValidationException

log = logging.getLogger(__name__)


class ReportType(str, Enum):
    NUDITY = 'nudity'
    PROFANITY = 'profanity'
    ILLEGAL = 'illegal'
    SPAM = 'spam'
    IMPERSONATION = 'impersonation'


@dataclass
class Report:
    """NIP-56 reporting event."""

    content: Optional[str] = None
    pubkey: Optional[str] = None
    reported_pubkey: Optional[str] = None
    note_id: Optional[str] = None
    report_type: Optional[ReportType] = None
    victim_pubkey: Optional[str] = None

    def to_event(self) -> Event:
        if self.reported_pubkey is None:
            raise NIPValidationException(
                "Reports require the pubkey of the user being reported"
            )
        if self.report_type is None or not isinstance(self.report_type, ReportType):
            raise NIPValidationException("Reports require a valid report type")
        e = Event(content=self.content, pubkey=self.pubkey, kind=EventKind.REPORT)
        if self.note_id:
            e.tags.append(["e", self.note_id, self.report_type])
            e.tags.append(["p", self.reported_pubkey])
        else:
            e.tags.append(["p", self.reported_pubkey, self.report_type])
            if self.victim_pubkey:
                e.tags.append(["p", self.victim_pubkey])
        e.compute_id()
        return e



import json
import logging
import time
from dataclasses import dataclass
from typing import Optional

from tornado import gen
from tornado.ioloop import IOLoop

from .base_relay import RelayPolicy
from .event import Event
from .exception import RelayException
from .filters import FiltersList
from .message_pool import MessagePool
from .relay import Relay

log = logging.getLogger(__name__)


@dataclass
class RelayManager:
    """RelayManager.

    :param error_threshold: When set, error_threshold on each relay is overwritten
    :param timeout:  When set, timeout on each relay is overwritten
    """

    error_threshold: Optional[int] = None
    timeout: Optional[float] = None

    def __post_init__(self):
        self.relays: dict[str, Relay] = {}
        self.message_pool: MessagePool = MessagePool()
        self.io_loop: IOLoop = IOLoop.current()

    def add_relay(
        self,
        url: str,
        policy: RelayPolicy = RelayPolicy(),
        timeout=2,
        close_on_eose: bool = True,
        message_callback=None,
        get_metadata=True,
    ):

        relay = Relay(
            url,
            self.message_pool,
            self.io_loop,
            policy,
            timeout=timeout,
            close_on_eose=close_on_eose,
            message_callback=message_callback,
        )

        if self.error_threshold is not None:
            relay.error_threshold = self.error_threshold
        if self.timeout is not None:
            relay.timeout = self.timeout
        if get_metadata:
            relay.update_metadata()
        self.relays[url] = relay

    def remove_relay(self, url: str):
        if url in self.relays:
            relay = self.relays.pop(url)
            relay.close()

    def remove_closed_relays(self):
        for url, connected in self.connection_statuses.items():
            if not connected:
                log.info(f'{url} is not connected... removing relay.')
                self.remove_relay(url=url)

    def add_subscription_on_relay(self, url: str, id: str, filters: FiltersList):
        if url in self.relays:
            relay = self.relays[url]
            if not relay.policy.should_read:
                raise RelayException(
                    f"Could not send request: {url} " f"is not configured to read from"
                )
            relay.add_subscription(id, filters)

    @gen.coroutine
    def prepare_relays(self):
        futures_timeout = []
        futures = []
        relays = []
        for relay in self.relays.values():
            if relay.policy.should_read:
                # yield relay.connect()
                timeout = relay.timeout
                relays.append(relay)
                if timeout > 0:
                    future = gen.with_timeout(
                        self.io_loop.time() + timeout, relay.connect()
                    )
                    futures_timeout.append(future)
                else:
                    futures.append(relay.connect())

        if len(futures_timeout) > 0:
            for i, future in enumerate(futures_timeout):
                try:
                    yield future
                except gen.TimeoutError:
                    log.warning(
                        f"Connection to WebSocket client {relays[i].url} timed out"
                    )
        elif len(futures) > 0:
            yield gen.multi(futures)
        raise gen.Return(relays)

    def add_subscription_on_all_relays(self, id: str, filters: FiltersList):
        for relay in self.relays.values():
            if relay.policy.should_read:
                relay.add_subscription(id, filters)

    def run_sync(self):
        self.io_loop.run_sync(lambda: self.prepare_relays())

    def close_subscription_on_relay(self, url: str, id: str):
        if url in self.relays:
            relay = self.relays[url]
            relay.close_subscription(id)
            relay.publish(json.dumps(["CLOSE", id]))
        else:
            raise RelayException(f"Invalid relay url: no connection to {url}")

    def close_subscription_on_all_relays(self, id: str):
        for relay in self.relays.values():
            relay.close_subscription(id)
            relay.publish(json.dumps(["CLOSE", id]))

    def close_all_relay_connections(self):
        for url in self.relays:
            relay = self.relays[url]
            relay.close()

    def open_connections(self, ssl_options: dict = None):
        for relay in self.relays.values():
            if not relay.is_connected:
                self.io_loop.add_callback(relay.start)
                self.io_loop.start()
        time.sleep(2)
        self.remove_closed_relays()

    def close_connections(self):
        for relay in self.relays.values():
            relay.close()

        return any(self.connection_statuses.values())

    @property
    def connection_statuses(self) -> dict:
        """gets the url and connection statuses of relays
        Returns:
            dict: bool of connection statuses
        """
        statuses = [relay.is_connected for relay in self.relays.values()]
        return dict(zip(self.relays.keys(), statuses))

    def publish_message(self, message: str):
        for relay in self.relays.values():
            if relay.policy.should_write:
                relay.publish(message)

    def publish_event(self, event: Event):
        """Verifies that the Event is publishable before submitting it to relays."""
        if event.sig is None:
            raise RelayException(f"Could not publish {event.id}: must be signed")

        if not event.verify():
            raise RelayException(
                f"Could not publish {event.id}: failed to verify signature {event.sig}"
            )

        self.publish_message(event.to_message())

    def get_relay_information(self, update_metadata=False):
        ret = {}
        for url in self.relays:
            if update_metadata:
                self.relays[url].update_metadata()
            ret[url] = self.relays[url].metadata
        return ret



"""Forked from https://github.com/jeffthibault/python-nostr.git."""
import json
from dataclasses import dataclass
from queue import Queue
from threading import Lock
from typing import List, Optional

from .event import Event
from .message_type import RelayMessageType


@dataclass
class EventMessage:
    event: str
    subscription_id: str
    url: str

    def __repr__(self):
        return f'EventMessage({self.url}: kind {str(self.event.kind)})'


@dataclass
class NoticeMessage:
    content: str
    url: str

    def __repr__(self):
        return f'Notice({self.url}: {self.content})'


@dataclass
class OKMessage:
    event_id: str
    ok: str
    message: str
    url: str

    def __repr__(self):
        return f'OK({self.url}: {self.event_id} {self.ok} {self.message})'


@dataclass
class EndOfStoredEventsMessage:
    subscription_id: str
    url: str

    def __repr__(self):
        return f'EOSE({self.url})'


class MessagePool:
    def __init__(self, first_response_only: bool = False):
        self.first_response_only = first_response_only
        self.events: Queue[EventMessage] = Queue()
        self.notices: Queue[NoticeMessage] = Queue()
        self.eose_notices: Queue[EndOfStoredEventsMessage] = Queue()
        self.ok_notices: Queue[OKMessage] = Queue()
        self._unique_objects: set = set()
        self.lock: Lock = Lock()

    def add_message(self, message: str, url: str):
        self._process_message(message, url)

    def get_all_events(self):
        events = []
        while self.has_events():
            events.append(self.get_event())
        return events

    def get_all_notices(self):
        notices = []
        while self.has_notices():
            notices.append(self.get_notice())
        return notices

    def get_all_eose(self):
        eose = []
        while self.has_eose_notices():
            eose.append(self.get_eose_notice())
        return eose

    def get_all_ok(self):
        ok = []
        while self.has_ok_notices():
            ok.append(self.get_ok_notice())
        return ok

    def get_all(self):
        results = {"events": [], "notices": [], "eose": [], "ok": []}
        results["events"] = self.get_all_events()
        results["notices"] = self.get_all_notices()
        results["eose"] = self.get_all_eose()
        results["ok"] = self.get_all_ok()
        return results

    def get_event(self):
        return self.events.get()

    def get_notice(self):
        return self.notices.get()

    def get_eose_notice(self):
        return self.eose_notices.get()

    def get_ok_notice(self):
        return self.ok_notices.get()

    def has_events(self):
        return self.events.qsize() > 0

    def has_notices(self):
        return self.notices.qsize() > 0

    def has_eose_notices(self):
        return self.eose_notices.qsize() > 0

    def has_ok_notices(self):
        return self.ok_notices.qsize() > 0

    def _process_message(self, message: str, url: str):
        message_json = json.loads(message)
        message_type = message_json[0]
        if message_type == RelayMessageType.EVENT:
            subscription_id = message_json[1]
            e = message_json[2]
            event = Event.from_dict(e)
            with self.lock:
                if self.first_response_only:
                    object_id = event.id
                else:
                    object_id = f'{event.id}:{url}'
                if object_id not in self._unique_objects:
                    self.events.put(EventMessage(event, subscription_id, url))
                    self._unique_objects.add(event.id)
        elif message_type == RelayMessageType.NOTICE:
            self.notices.put(NoticeMessage(message_json[1], url))
        elif message_type == RelayMessageType.END_OF_STORED_EVENTS:
            self.eose_notices.put(EndOfStoredEventsMessage(message_json[1], url))
        elif message_type == RelayMessageType.OK:
            self.ok_notices.put(
                OKMessage(message_json[1], message_json[2], message_json[3], url)
            )

    def __repr__(self):
        return (
            f'Pool(events({self.events.qsize()}) notices({self.notices.qsize()}) '
            f'eose({self.eose_notices.qsize()}) ok({self.ok_notices.qsize()}))'
        )


@dataclass
class EventMessageStore:
    eventMessages: Optional[List[EventMessage]] = None

    def add_event(self, event):
        if self.eventMessages is None:
            self.eventMessages = []
        if isinstance(event, list):
            self.eventMessages += event
        else:
            self.eventMessages.append(event)

    def get_newest_event(self):
        if not self.eventMessages:
            return None
        return max(self.eventMessages, key=lambda x: x.event.date_time())

    def get_events_by_url(self, url):
        return [event for event in self.eventMessages if event.url == url]

    def get_events_by_id(self, subscription_id):
        return [
            event
            for event in self.eventMessages
            if event.subscription_id == subscription_id
        ]

    def __repr__(self):
        if not self.eventMessages:
            return 'EventMessageStore()'
        return f'EventMessageStore({len(self.eventMessages)} events)'



import binascii
import datetime
import logging

import requests
import tlv8

from . import bech32

log = logging.getLogger(__name__)


def bech32_decode(bech32_str: str):
    """Loads bytes from its bech32/nsec form."""
    hrp, data, spec = bech32.bech32_decode(bech32_str)
    raw_secret = bech32.convertbits(data, 5, 8)
    if raw_secret[-1] != 0x0:
        return bytes(raw_secret)
    else:
        return bytes(raw_secret[:-1])


def bech32_encode(raw_bytes: bytes, prefix: str) -> str:
    converted_bits = bech32.convertbits(raw_bytes, 8, 5)
    return bech32.bech32_encode(prefix, converted_bits, bech32.Encoding.BECH32)


def split_nip05(nip05):
    if nip05 is None:
        return None, None
    if "@" in nip05:
        name, url = nip05.lower().split("@")
    else:
        name = "_"
        url = nip05.lower()
    return name, url


def get_nip05_response(name, url):
    if url is None:
        return {}
    request_url = f"https://{url}/.well-known/nostr.json?name={name}"
    try:
        response = requests.get(
            request_url, headers={'User-Agent': 'pynostr'}, timeout=5
        )

        response.raise_for_status()

        return response.json()
    except requests.exceptions.Timeout:
        # Handle a timeout error
        log.warning("Request timed out. Please try again later.")

    except requests.exceptions.HTTPError as err:
        # Handle an HTTP error
        log.warning(f"HTTP error occurred: {err}")

    except requests.exceptions.RequestException as err:
        # Handle any other request exception
        log.warning(f"An error occurred: {err}")
    return None


def check_nip05(nip05_response: dict, name: str):
    if len(nip05_response) == 0:
        return
    if "names" in nip05_response:
        if name in nip05_response["names"]:
            pubkey = nip05_response["names"][name]
        else:
            pubkey = None
    else:
        pubkey = None
    if "relays" in nip05_response and pubkey is not None:
        if pubkey in nip05_response["relays"]:
            relays = nip05_response["relays"][pubkey]
        else:
            relays = None
    else:
        relays = None
    return pubkey, relays


def extract_nip05(nip05: str):
    name, url = split_nip05(nip05)
    nip05_response = get_nip05_response(name, url)
    return check_nip05(nip05_response, name)


def get_relay_information(url: str, timeout: float = 2, add_url: bool = True):
    headers = {'Accept': 'application/nostr+json', 'User-Agent': 'pynostr'}
    if "wss" in url:
        metadata_uri = url.replace("wss", "https")
    elif "ws" in url:
        metadata_uri = url.replace("ws", "http")
    else:
        raise Exception(f"{url} is not a websocket url")
    try:
        response = requests.get(metadata_uri, headers=headers, timeout=timeout)

        response.raise_for_status()

        metadata = response.json()
        if add_url:
            metadata["url"] = url
        return metadata
    except requests.exceptions.Timeout:
        # Handle a timeout error
        log.warning("Request timed out. Please try again later.")

    except requests.exceptions.HTTPError as err:
        # Handle an HTTP error
        log.warning(f"HTTP error occurred: {err}")

    except requests.exceptions.RequestException as err:
        # Handle any other request exception
        log.warning(f"An error occurred: {err}")


def nprofile_decode(nprofile: str):
    decode = bytes(bech32_decode(nprofile))
    pubkey = None
    relays = []
    for data in tlv8.deep_decode(decode):
        if data.type_id == 0:
            pubkey = binascii.hexlify(data.data).decode()
        elif data.type_id == 1:
            relays.append(data.data.decode())
    return pubkey, relays


def nprofile_encode(pubkey: str, relays: [str]):
    structure = [
        tlv8.Entry(0, binascii.unhexlify(pubkey)),
    ]
    for relay in relays:
        structure.append(tlv8.Entry(1, relay.encode()))

    bytes_data = tlv8.encode(structure)

    bytes_data = b''
    for entry in structure:
        bytes_data += entry.encode()
    return bech32_encode(bytes_data, "nprofile")


def get_timestamp(days=0, seconds=0, minutes=0, hours=0, weeks=0):
    now = datetime.datetime.now()
    date = now - datetime.timedelta(
        days=days, seconds=seconds, minutes=minutes, hours=hours, weeks=weeks
    )
    return int(date.timestamp())


def get_relay_list(relay_type="online", nip=None, timeout=5):
    """Uses the API from https://api.nostr.watch/

    :param relay_type: can be online, offline or nip
    :param nip: is used when relay_type is set to nip
    """
    headers = {'User-Agent': 'pynostr'}
    url = f"https://api.nostr.watch/v1/{relay_type}"
    if nip is not None:
        url += f"/{nip}"
    try:
        response = requests.get(url, headers=headers, timeout=timeout)

        response.raise_for_status()

        data = response.json()
        return data
    except requests.exceptions.Timeout:
        # Handle a timeout error
        log.warning("Request timed out. Please try again later.")

    except requests.exceptions.HTTPError as err:
        # Handle an HTTP error
        log.warning(f"HTTP error occurred: {err}")

    except requests.exceptions.RequestException as err:
        # Handle any other request exception
        log.warning(f"An error occurred: {err}")



import json
import logging
from dataclasses import dataclass
from typing import List, Optional

from .event import Event, EventKind
from .key import PublicKey
from .utils import extract_nip05

log = logging.getLogger(__name__)


@dataclass
class Metadata:
    pubkey: Optional[str] = None
    name: Optional[str] = None
    about: Optional[str] = None
    nip05: Optional[str] = None
    picture: Optional[str] = None
    banner: Optional[str] = None
    lud16: Optional[str] = None
    lud06: Optional[str] = None
    username: Optional[str] = None
    display_name: Optional[str] = None
    website: Optional[str] = None
    addional: Optional[dict] = None
    event: Optional[Event] = None
    relays: Optional[List[str]] = None

    @classmethod
    def from_dict(cls, msg: dict) -> 'Metadata':
        # "id" is ignore, as it will be computed from the contents
        p = Metadata()
        if "name" in msg:
            p.name = msg.pop("name")
        if "about" in msg:
            p.about = msg.pop("about")
        if "nip05" in msg:
            p.nip05 = msg.pop("nip05")
        if "picture" in msg:
            p.picture = msg.pop("picture")
        if "banner" in msg:
            p.banner = msg.pop("banner")
        if "lud16" in msg:
            p.lud16 = msg.pop("lud16")
        if "lud06" in msg:
            p.lud06 = msg.pop("lud06")
        if "username" in msg:
            p.username = msg.pop("username")
        if "display_name" in msg:
            p.display_name = msg.pop("display_name")
        if "website" in msg:
            p.website = msg.pop("website")
        if len(msg) > 0:
            p.addional = msg
        return p

    @classmethod
    def from_event(cls, event: Event) -> 'Metadata':
        if event.kind != EventKind.SET_METADATA:
            return None
        metadata = json.loads(event.content)
        m = Metadata.from_dict(metadata)
        m.pubkey = event.pubkey
        m.event = event
        return m

    @classmethod
    def from_nip05(cls, nip05: str) -> 'Metadata':
        m = Metadata()
        m.nip05 = nip05
        try:
            pubkey, relays = extract_nip05(nip05)
        except Exception:
            return m
        if pubkey is not None:
            m.pubkey = pubkey
        if relays is not None:
            m.relays = relays
        return m

    @classmethod
    def from_npub(cls, npub: str) -> 'Metadata':
        m = Metadata()
        m.pubkey = PublicKey.from_npub(npub).hex()
        return m

    def validate_nip05(self):
        if self.nip05 is None or self.pubkey is None:
            return False
        try:
            pubkey, relays = extract_nip05(self.nip05)
        except Exception:
            return False
        if pubkey is not None:
            if pubkey == self.pubkey:
                return True

        return False



import unittest

from pynostr.delegation import Delegation
from pynostr.event import Event, EventKind
from pynostr.key import PrivateKey


class TestDelegation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.identity_pk = PrivateKey()
        cls.delegatee_pk = PrivateKey()

    def test_event_signing(self):
        delegation = Delegation(
            delegator_pubkey=self.identity_pk.public_key.hex(),
            delegatee_pubkey=self.delegatee_pk.public_key.hex(),
            event_kind=EventKind.TEXT_NOTE,
            duration_secs=30 * 24 * 60 * 60,
        )

        self.identity_pk.sign_delegation(delegation)

        event = Event(
            "Hello, NIP-26!",
            tags=[delegation.get_tag()],
        )
        event.sign(self.delegatee_pk.hex())
        self.assertTrue(event.sig is not None)



#!/usr/bin/env python

import uuid

import tornado.ioloop
from rich.console import Console
from rich.table import Table
from tornado import gen

from pynostr.base_relay import RelayPolicy
from pynostr.event import EventKind
from pynostr.filters import Filters, FiltersList
from pynostr.key import PublicKey
from pynostr.message_pool import MessagePool
from pynostr.relay import Relay
from pynostr.utils import extract_nip05

if __name__ == "__main__":

    console = Console()

    input_str = input("author (npub or nip05): ")
    recipient = ""
    if "@" in input_str:
        nip05 = extract_nip05(input_str)
        author = PublicKey.from_hex(nip05[0])
    elif "npub" in input_str:
        author = PublicKey.from_hex(input_str)
    if author != "":
        print(f"author is set to {author.bech32()}")
    else:
        raise Exception("reciever not valed")
    relay_url = input("relay: ")

    filters = FiltersList(
        [Filters(authors=[author.hex()], kinds=[EventKind.TEXT_NOTE], limit=100)]
    )

    subscription_id = uuid.uuid1().hex
    io_loop = tornado.ioloop.IOLoop.current()
    message_pool = MessagePool(first_response_only=False)
    policy = RelayPolicy()
    r = Relay(relay_url, message_pool, io_loop, policy, timeout=3)

    r.add_subscription(subscription_id, filters)

    try:
        io_loop.run_sync(r.connect)
    except gen.Return:
        pass
    io_loop.stop()

    event_msgs = message_pool.get_all_events()
    print(f"{r.url} returned {len(event_msgs)} TEXT_NOTEs from {input_str}.")

    table = Table("date", "content")
    for event_msg in event_msgs[::-1]:
        table.add_row(str(event_msg.event.date_time()), event_msg.event.content)
    console.print(table)



import unittest

from pynostr.encrypted_dm import EncryptedDirectMessage
from pynostr.event import Event
from pynostr.exception import NIPValidationException
from pynostr.key import PrivateKey


class TestEncryptedDirectMessage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.sender_pk = PrivateKey.from_hex(
            "29307c4354b7d9d311d2cec4878c0de56c93a921d300273c19577e9004de3c9f"
        )
        cls.sender_pubkey = cls.sender_pk.public_key.hex()
        cls.recipient_pk = PrivateKey.from_hex(
            "4138d1b6dde34f81c38cef2630429e85847dd5b70508e37f53c844f66f19f983"
        )
        cls.recipient_pubkey = cls.recipient_pk.public_key.hex()

    def test_recipient_p_tag(self):
        """Should generate recipient 'p' tag."""
        dm = EncryptedDirectMessage()
        dm.encrypt(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content="Secret message!",
        )
        self.assertIn(['p', self.recipient_pubkey], dm.to_event().tags)

    def test_event_e_tag(self):
        """May generate event 'e' tag."""
        dm = EncryptedDirectMessage()
        dm.reference_event_id = "ref_event_id"
        dm.encrypt(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content="Secret message!",
        )
        self.assertIn(['p', self.recipient_pubkey], dm.to_event().tags)
        self.assertIn(['e', "ref_event_id"], dm.to_event().tags)

    def test_NIP_validation(self):
        dm = EncryptedDirectMessage()
        with self.assertRaisesRegex(
            NIPValidationException, "Encrypted message is missing"
        ):
            dm.to_event()
        dm.encrypt(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content="Secret message!",
        )
        dm.recipient_pubkey = None
        with self.assertRaisesRegex(
            NIPValidationException, "recipient_pubkey is missing"
        ):
            dm.to_event()

    def test_encrypt_dm(self):
        """Should encrypt a DM and populate its `content` field with ciphertext that
        either party can decrypt."""
        message1 = "Test"
        message2 = "Test2"
        message3 = "Test3"
        message4 = "Test4"

        dm1 = EncryptedDirectMessage()
        dm1.encrypt(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content=message1,
        )
        dm2 = EncryptedDirectMessage()
        dm2.encrypt(
            self.recipient_pk.hex(),
            recipient_pubkey=self.sender_pubkey,
            cleartext_content=message2,
        )
        dm3 = EncryptedDirectMessage()
        dm3.encrypt(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content=message3,
        )
        dm4 = EncryptedDirectMessage()
        dm4.encrypt(
            self.recipient_pk.hex(),
            recipient_pubkey=self.sender_pubkey,
            cleartext_content=message4,
        )
        dm1_event = dm1.to_event()
        dm2_event = dm2.to_event()
        dm3_event = dm3.to_event()
        dm4_event = dm4.to_event()

        # After encrypting, the content field should now be populated
        self.assertTrue(dm1_event.content is not None)

        # Sender should be able to decrypt
        decm1_sender = self.sender_pk.decrypt_message(
            encoded_message=dm1_event.content, public_key_hex=self.recipient_pubkey
        )
        decm2_sender = self.sender_pk.decrypt_message(
            encoded_message=dm2_event.content, public_key_hex=self.recipient_pubkey
        )
        decm3_sender = self.sender_pk.decrypt_message(
            encoded_message=dm3_event.content, public_key_hex=self.recipient_pubkey
        )
        decm4_sender = self.sender_pk.decrypt_message(
            encoded_message=dm4_event.content, public_key_hex=self.recipient_pubkey
        )

        decm1_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm1_event.content, public_key_hex=self.sender_pubkey
        )
        decm2_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm2_event.content, public_key_hex=self.sender_pubkey
        )
        decm3_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm3_event.content, public_key_hex=self.sender_pubkey
        )
        decm4_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm4_event.content, public_key_hex=self.sender_pubkey
        )
        self.assertEqual(message1, decm1_receiver)
        self.assertEqual(message1, decm1_sender)
        self.assertEqual(message2, decm2_receiver)
        self.assertEqual(message2, decm2_sender)
        self.assertEqual(message3, decm3_receiver)
        self.assertEqual(message3, decm3_sender)
        self.assertEqual(message4, decm4_receiver)
        self.assertEqual(message4, decm4_sender)

    def test_decrypt_dm(self):
        message1 = "Test"
        message2 = "Test2"
        message3 = "Test3"
        message4 = "Test4"
        encrypt1 = "VOqWLiW4wv8+fDsNC00a1w==?iv=LSIH1sk13Mw09PV8Z80sag=="
        encrypt2 = "abZBRLPta8888xDkg6pUWA==?iv=Gj5KOUbFqREhSdbMENRKEg=="
        encrypt3 = "w2AsXNN0EysjG6/E/GZWPg==?iv=3c7qsPxSOckGeqjjpwQQ4A=="
        encrypt4 = "nBfP5P2GEEOlfNYMoxADDg==?iv=VsFd7eE8BfoyDJpfQ7kjhQ=="

        dm1 = EncryptedDirectMessage(encrypted_message=encrypt1)
        dm2 = EncryptedDirectMessage(encrypted_message=encrypt2)
        dm3 = EncryptedDirectMessage(encrypted_message=encrypt3)
        dm4 = EncryptedDirectMessage(encrypted_message=encrypt4)
        dm1.decrypt(self.sender_pk.hex(), public_key_hex=self.recipient_pubkey)
        dm2.decrypt(self.sender_pk.hex(), public_key_hex=self.recipient_pubkey)
        dm3.decrypt(self.sender_pk.hex(), public_key_hex=self.recipient_pubkey)
        dm4.decrypt(self.sender_pk.hex(), public_key_hex=self.recipient_pubkey)

        self.assertEqual(dm1.cleartext_content, message1)
        self.assertEqual(dm2.cleartext_content, message2)
        self.assertEqual(dm3.cleartext_content, message3)
        self.assertEqual(dm4.cleartext_content, message4)

        dm1.cleartext_content = None
        dm2.cleartext_content = None
        dm3.cleartext_content = None
        dm4.cleartext_content = None

        dm1.decrypt(self.recipient_pk.hex(), public_key_hex=self.sender_pubkey)
        dm2.decrypt(self.recipient_pk.hex(), public_key_hex=self.sender_pubkey)
        dm3.decrypt(self.recipient_pk.hex(), public_key_hex=self.sender_pubkey)
        dm4.decrypt(self.recipient_pk.hex(), public_key_hex=self.sender_pubkey)

        self.assertEqual(dm1.cleartext_content, message1)
        self.assertEqual(dm2.cleartext_content, message2)
        self.assertEqual(dm3.cleartext_content, message3)
        self.assertEqual(dm4.cleartext_content, message4)

    def test_sign_encrypts_dm(self):
        """`sign` should encrypt a DM that hasn't been encrypted yet."""
        dm = EncryptedDirectMessage()
        self.assertTrue(dm.cleartext_content is None)
        self.assertTrue(dm.encrypted_message is None)
        dm.encrypt(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content="Some DM message",
        )
        dm_event = dm.to_event()
        dm_event.sign(self.sender_pk.hex())

        self.assertTrue(dm_event.content is not None)

    def test_shared_secret(self):
        sender_pk = "29307c4354b7d9d311d2cec4878c0de56c93a921d300273c19577e9004de3c9f"
        recipient_pk = (
            "4138d1b6dde34f81c38cef2630429e85847dd5b70508e37f53c844f66f19f983"
        )

        sender_pubkey = PrivateKey.from_hex(sender_pk).public_key.hex()
        recipient_pubkey = PrivateKey.from_hex(recipient_pk).public_key.hex()

        shared_secret1 = PrivateKey.from_hex(sender_pk).compute_shared_secret(
            recipient_pubkey
        )
        shared_secret2 = PrivateKey.from_hex(recipient_pk).compute_shared_secret(
            sender_pubkey
        )
        self.assertEqual(shared_secret1, shared_secret2)

    def test_decrypt_event(self):
        dm1 = Event.from_dict(
            {
                "id": "46c76ec67d03babdb840254b3667585143cc499497b0a6a40aedc9ce2de416"
                "70",
                "pubkey": "f3c25355c29f64ea8e9b4e11b583ac0a7d0d8235f156cffec2b73e5756a"
                "ab206",
                "created_at": 1674819397,
                "kind": 4,
                "tags": [
                    [
                        "p",
                        "a1db8e8b047e1350958a55e0a853151d0e1f685fa5cf3772e01bccc5aa5c"
                        "b2eb",
                    ]
                ],
                "content": "VOqWLiW4wv8+fDsNC00a1w==?iv=LSIH1sk13Mw09PV8Z80sag==",
                "sig": "982e825d32e78bf695a2511e14d2958e0d480bfa338e12307e5eb7ecad3b"
                "d86d53c2d508df9083a0fc25b44c44ae4c622f7405bbc2ca723c5dbc98da48214f16",
            }
        )
        self.assertTrue(dm1.verify())



import json
import logging
from dataclasses import dataclass
from queue import Queue
from threading import Lock
from typing import Dict

from .event import Event
from .filters import FiltersList
from .message_pool import MessagePool
from .message_type import RelayMessageType
from .request import Request
from .subscription import Subscription
from .utils import get_relay_information

log = logging.getLogger(__name__)


@dataclass
class RelayPolicy:
    should_read: bool = True
    should_write: bool = True

    def to_json_object(self) -> Dict[str, bool]:
        return {"read": self.should_read, "write": self.should_write}


@dataclass
class RelayProxyConnectionConfig:
    host: str
    port: int
    type: str


class BaseRelay:
    def __init__(
        self,
        url: str,
        message_pool: MessagePool = MessagePool(),
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: RelayProxyConnectionConfig = None,
        timeout: float = 2.0,
        close_on_eose: bool = True,
        message_callback=None,
    ) -> None:
        self.url = url
        self.message_pool = message_pool
        self.policy = policy
        self.ssl_options = ssl_options
        self.proxy_config = proxy_config
        self.timeout = timeout
        self.close_on_eose = close_on_eose
        self.lock: Lock = Lock()
        self.ws = None
        self.metadata = None
        self.subscriptions: dict[str, Subscription] = {}
        self.connected: bool = False
        self.error_counter: int = 0
        self.error_threshold: int = 10
        self.num_sent_events: int = 0
        self.request: str = ""
        self.message_callback = message_callback
        self.outgoing_messages = Queue()

    def __repr__(self):
        return json.dumps(self.to_json_object(), indent=2)

    def to_json_object(self) -> dict:
        return {
            "url": self.url,
            "policy": self.policy.to_json_object(),
            "subscriptions": [
                subscription.to_json_object()
                for subscription in self.subscriptions.values()
            ],
            "request": self.request,
        }

    def update_metadata(self, timeout: float = None) -> None:
        if timeout is None:
            timeout = self.timeout
        self.metadata = get_relay_information(self.url, timeout=timeout, add_url=False)

    def check_nip(self, nip: int) -> bool:
        if self.metadata is None:
            self.update_metadata()
        if self.metadata is None:
            return False
        if "supported_nips" not in self.metadata:
            return False
        return nip in self.metadata["supported_nips"]

    def add_subscription(self, id, filters: FiltersList):
        with self.lock:
            self.subscriptions[id] = Subscription(id, filters)
            self.request = Request(id, filters).to_message()

    def close_subscription(self, id: str) -> None:
        with self.lock:
            self.subscriptions.pop(id, None)

    def update_subscription(self, id: str, filters: FiltersList) -> None:
        with self.lock:
            subscription = self.subscriptions[id]
            subscription.filters = filters
            self.request = Request(id, filters).to_message()

    def _on_message(self, message):
        if self._is_valid_message(message):
            message_json = json.loads(message)
            if self.message_callback is not None:
                self.message_callback(message_json)
            message_type = message_json[0]
            if message_type == RelayMessageType.EVENT:
                # event = Event.from_dict(message_json[2])
                # print(event.to_message())
                self.message_pool.add_message(message, self.url)
            elif message_type == RelayMessageType.END_OF_STORED_EVENTS:
                self._eose_received()
                self.message_pool.add_message(message, self.url)
            elif message_type == RelayMessageType.OK:
                self.message_pool.add_message(message, self.url)
            elif message_type == RelayMessageType.AUTH:
                print(message)

    def publish(self, message: str):
        self.outgoing_messages.put(message)

    def _eose_received(self):
        return

    def _is_valid_message(self, message: str) -> bool:
        if message is None:
            return False
        message = message.strip("\n")
        if not message or message[0] != "[" or message[-1] != "]":
            return False

        message_json = json.loads(message)
        message_type = message_json[0]
        if not RelayMessageType.is_valid(message_type):
            return False
        if message_type == RelayMessageType.EVENT:
            if not len(message_json) == 3:
                return False

            subscription_id = message_json[1]
            with self.lock:
                if subscription_id not in self.subscriptions:
                    return False

            event = Event.from_dict(message_json[2])
            if not event.verify():
                return False

            with self.lock:
                subscription = self.subscriptions[subscription_id]

            if subscription.filtersList and not subscription.filtersList.match(event):
                return False
        elif message_type == RelayMessageType.OK:
            if not len(message_json) == 4:
                return False
            if message_json[2] in ["true", "false"]:
                return True
            elif isinstance(message_json[2], bool):
                return True
            else:
                return False
        return True



import unittest

from pynostr.base_relay import BaseRelay
from pynostr.message_pool import MessagePool


class TestBaseRelay(unittest.TestCase):
    def test_valid_message(self):
        message_pool = MessagePool()
        b = BaseRelay("wss://test.test", message_pool)
        message = (
            '["OK","fafb7406c6245e3a259b51aa6eccb9433fd765af2302ea80738f4ad8ce2'
            '2e55e",false,"blocked: not on white-list"]'
        )
        self.assertTrue(b._is_valid_message(message))

        b._on_message(message)
        self.assertTrue(message_pool.has_ok_notices())



"""Forked from https://github.com/jeffthibault/python-nostr.git."""
import time
import unittest

from pynostr.event import Event
from pynostr.key import PrivateKey


class TestEvent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.sender_pk = PrivateKey()
        cls.sender_pubkey = cls.sender_pk.public_key.hex()

    def test_note_event(self):
        pk1 = PrivateKey.from_hex(
            "964b29795d621cdacf05fd94fb23206c88742db1fa50b34d7545f3a2221d8124"
        )
        event = Event("Hello Nostr!", pk1.public_key.hex())
        event.created_at = 1671406583
        event.compute_id()
        event.sign(pk1.hex())
        self.assertEqual(
            "23411895658d374ec922adf774a70172290b2c738ae67815bd8945e5d8fff3bb", event.id
        )
        self.assertTrue(event.verify())

    def test_event_default_time(self):
        public_key = PrivateKey().public_key.hex()
        event1 = Event(pubkey=public_key, content='test event')
        time.sleep(1.5)
        event2 = Event(pubkey=public_key, content='test event')
        self.assertTrue(event1.created_at < event2.created_at)

    def test_add_event_ref(self):
        some_event_id = "some_event_id"
        event = Event(content="Adding an 'e' tag")
        event.add_event_ref(some_event_id)
        self.assertTrue(['e', some_event_id] in event.tags)
        self.assertEqual(event.get_tag_count('e'), 1)
        self.assertEqual(event.get_tag_count('p'), 0)
        self.assertEqual(event.get_tag_list('e'), [some_event_id])
        self.assertEqual(event.get_tag_list('p'), [])
        self.assertEqual(event.get_tag_types(), ['e'])
        self.assertEqual(event.get_tag_dict(), {"e": [some_event_id]})
        self.assertTrue(event.has_event_ref(some_event_id))

    def test_add_pubkey_ref(self):
        some_pubkey = "some_pubkey"
        event = Event(content="Adding a 'p' tag")
        event.add_pubkey_ref(some_pubkey)
        self.assertTrue(['p', some_pubkey] in event.tags)
        self.assertEqual(event.get_tag_count('p'), 1)
        self.assertEqual(event.get_tag_count('e'), 0)
        self.assertEqual(event.get_tag_list('p'), [some_pubkey])
        self.assertEqual(event.get_tag_list('e'), [])
        self.assertEqual(event.get_tag_types(), ['p'])
        self.assertEqual(event.get_tag_dict(), {"p": [some_pubkey]})
        self.assertTrue(event.has_pubkey_ref(some_pubkey))

    def test_sign_event_is_valid(self):
        """Sign should create a signature that can be verified against Event.id."""
        event = Event(content="Hello, world!")
        event.sign(self.sender_pk.hex())
        self.assertTrue(event.verify())

    def test_sign_event_adds_pubkey(self):
        """Sign should add the sender's pubkey if not already specified."""
        event = Event(content="Hello, world!")

        # The event's public_key hasn't been specified yet
        self.assertTrue(event.pubkey is None)
        event.sign(self.sender_pk.hex())

        # PrivateKey.sign() should have populated public_key
        self.assertEqual(event.pubkey, self.sender_pubkey)

    def test_dict_roundtrip(self):
        """Conversion to dict and back result in same object."""
        event = Event(
            content='test event',
            created_at=12345678,
            kind=1,
        )
        event.add_pubkey_ref("some_pubkey")

        got = Event.from_dict(event.to_dict())
        self.assertEqual(got, event)



"""Forked from https://github.com/jeffthibault/python-nostr.git."""
import time
from threading import Thread
from typing import Union

from websocket import WebSocketApp

from .base_relay import BaseRelay, RelayPolicy, RelayProxyConnectionConfig
from .message_pool import MessagePool


class WebSocketRelay(BaseRelay):
    def __init__(
        self,
        url: str,
        message_pool: MessagePool,
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: Union[None, RelayProxyConnectionConfig] = None,
    ) -> None:
        super().__init__(url, message_pool, policy, ssl_options, proxy_config)
        self.ws: WebSocketApp = WebSocketApp(
            self.url,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
        )
        self._connection_thread: Thread = None

    @property
    def is_connected(self) -> bool:
        with self.lock:
            if (
                self._connection_thread is None
                or not self._connection_thread.is_alive()
            ):
                return False
            else:
                return True

    def close(self):
        if self.is_connected:
            self.ws.close()

    def connect(self, is_reconnect=False):
        if not self.is_connected:
            with self.lock:
                self._connection_thread = Thread(
                    target=self.ws.run_forever,
                    kwargs={
                        "sslopt": self.ssl_options,
                        "http_proxy_host": self.proxy_config.host
                        if self.proxy_config is not None
                        else None,
                        "http_proxy_port": self.proxy_config.port
                        if self.proxy_config is not None
                        else None,
                        "proxy_type": self.proxy_config.type
                        if self.proxy_config is not None
                        else None,
                    },
                    name=f"{self.url}-connection",
                )
                self._connection_thread.start()

            if not is_reconnect:
                Thread(
                    target=self.outgoing_messages_worker,
                    name=f"{self.url}-outgoing-messages-worker",
                    daemon=True,
                ).start()

            time.sleep(1)

    def outgoing_messages_worker(self):
        while True:
            if self.is_connected:
                message = self.outgoing_messages.get()
                try:
                    self.ws.send(message)
                    self.num_sent_events += 1
                except Exception:
                    self.outgoing_messages.put(message)

    def _on_open(self, class_obj):
        pass

    def _on_close(self, class_obj, status_code, message):
        self.error_counter = 0

    def _on_error(self, class_obj, error):
        self.error_counter += 1
        if self.error_counter > self.error_threshold:
            self.close()



import concurrent.futures
import datetime
import logging
import uuid

from rich.console import Console
from rich.table import Table

from pynostr.event import EventKind
from pynostr.filters import Filters, FiltersList
from pynostr.message_pool import EventMessageStore
from pynostr.relay_manager import RelayManager
from pynostr.utils import get_relay_information, get_relay_list, get_timestamp

log = logging.getLogger(__name__)

if __name__ == "__main__":

    console = Console()
    MAX_THREADS = 16

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch = logging.StreamHandler()
    ch.setLevel(4)
    ch.setFormatter(formatter)
    log.addHandler(ch)

    relay_list = get_relay_list()

    paid_relays = {}
    print(f"Checking {len(relay_list)} relays...")
    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_THREADS) as executor:
        res = executor.map(get_relay_information, relay_list)
        for relay_info in res:
            if relay_info is not None and "limitation" in relay_info:
                if "payment_required" in relay_info["limitation"]:
                    if relay_info["limitation"]["payment_required"]:
                        if (
                            "fees" not in relay_info
                            or len(relay_info["fees"]['admission']) == 0
                        ):
                            paid_relays[relay_info["url"]] = 0
                        elif "amount" in relay_info["fees"]['admission'][0]:
                            paid_relays[relay_info["url"]] = (
                                relay_info["fees"]['admission'][0]['amount'] / 1000
                            )
    print(f"Found {len(paid_relays)} relays with fee!")
    relay_manager = RelayManager(error_threshold=10, timeout=5)
    for url in paid_relays.keys():
        relay_manager.add_relay(url)

    events = EventMessageStore()
    events_by_relay = {}
    unix_timestamp = get_timestamp(days=7)
    now = datetime.datetime.utcnow()

    filters = FiltersList(
        [Filters(kinds=[EventKind.TEXT_NOTE], since=unix_timestamp, limit=500)]
    )
    subscription_id = uuid.uuid1().hex
    relay_manager.add_subscription_on_all_relays(subscription_id, filters)
    relay_manager.run_sync()

    event_messages = relay_manager.message_pool.get_all_events()
    events.add_event(event_messages)

    for url in paid_relays.keys():

        event_list = events.get_events_by_url(url)
        oldest_timestamp = now
        events_by_relay[url] = {
            "pubkey": [],
            "n_pubkey": 0,
            "admission": paid_relays[url],
            "oldest_event": None,
            "n_events": len(event_list),
        }
        for e in event_list:
            if e.event.date_time() < oldest_timestamp:
                oldest_timestamp = e.event.date_time()
            if e.event.pubkey not in events_by_relay[url]["pubkey"]:
                events_by_relay[url]["pubkey"].append(e.event.pubkey)
        events_by_relay[url]["oldest_event"] = oldest_timestamp
        events_by_relay[url]["n_pubkey"] = len(events_by_relay[url]['pubkey'])

    relay_list = sorted(events_by_relay.items(), key=lambda item: item)
    relay_list2 = sorted(relay_list, key=lambda item: item[1]["n_pubkey"])

    table = Table("url", "n_pubkey", "n_events", "age oldest_event", "fee")
    for r in relay_list2[::-1]:
        table.add_row(
            r[0],
            str(r[1]['n_pubkey']),
            str(r[1]['n_events']),
            f"{(now - r[1]['oldest_event']).total_seconds() / 60 / 60} h",
            str(r[1]['admission']),
        )
    console.print(table)



import logging

import click
import typer
from rich.console import Console
from rich.table import Table

from pynostr.key import PrivateKey, PublicKey
from pynostr.metadata import Metadata
from pynostr.utils import get_relay_information, nprofile_decode

log = logging.getLogger(__name__)
app = typer.Typer()
console = Console()

state = {"verbose": 3}


@app.command()
def keygen():
    """Creates a private and public key."""
    private_key = PrivateKey()
    public_key = private_key.public_key
    click.echo(f"Private key: {private_key.bech32()}")
    click.echo(f"Public key: {public_key.bech32()}")


@app.command()
def convert(identifier: str):
    """Converts npub key to hex."""
    if "npub" in identifier:
        public_key = PublicKey.from_npub(identifier)
    else:
        public_key = PublicKey.from_hex(identifier)
    click.echo(f"npub: {public_key.bech32()}")
    click.echo(f"hex: {public_key.hex()}")


@app.command()
def info(objects: str):
    """Basic info about object."""
    if "wss" in objects:
        relay_info = get_relay_information(objects)
        if relay_info is None:
            return
        table = Table("key", "value")
        for key, value in relay_info.items():
            if key in ["limitation"]:
                continue
            table.add_row(key, str(value))

        console.print(table)
        if "limitation" in relay_info:
            table = Table("key", "value")
            for key, value in relay_info["limitation"].items():
                table.add_row(key, str(value))
            console.print(table)
    elif "." in objects:
        m = Metadata.from_nip05(objects.lower())
        table = Table("key", "value")
        table.add_row("pubkey", str(m.pubkey))
        table.add_row("npub", str(PublicKey(m.pubkey).bech32()))
        console.print(table)
        if m.relays is not None:
            table = Table("relay")
            for url in m.relays:
                table.add_row(url)
            console.print(table)
    elif "nprofile" in objects:
        pubkey, relays = nprofile_decode(objects)
        table = Table("key", "value")
        table.add_row("pubkey", pubkey)
        table.add_row("npub", str(PublicKey(pubkey).bech32()))
        for i in range(len(relays)):
            table.add_row(f"relay{i}", relays[i])
        console.print(table)
    elif "npub" in objects:
        pubkey = PublicKey.from_npub(objects)
        table = Table("key", "value")
        table.add_row("pubkey", pubkey.hex())
        table.add_row("npub", str(pubkey.bech32()))
        console.print(table)


@app.callback()
def main(verbose: int = 3):
    """Python CLI for nostr, enjoy."""
    # Logging
    state["verbose"] = verbose
    log = logging.getLogger(__name__)
    verbosity = ["critical", "error", "warn", "info", "debug"][int(min(verbose, 4))]
    log.setLevel(getattr(logging, verbosity.upper()))
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, verbosity.upper()))
    ch.setFormatter(formatter)
    log.addHandler(ch)


if __name__ == "__main__":
    app()



#!/usr/bin/env python

import getpass
import uuid

import tornado.ioloop
from rich.console import Console
from tornado import gen

from pynostr.base_relay import RelayPolicy
from pynostr.encrypted_dm import EncryptedDirectMessage
from pynostr.event import Event, EventKind
from pynostr.filters import Filters, FiltersList
from pynostr.key import PrivateKey, PublicKey
from pynostr.message_pool import MessagePool
from pynostr.message_type import RelayMessageType
from pynostr.relay import Relay
from pynostr.utils import extract_nip05, get_timestamp


@gen.coroutine
def print_dm(message_json):
    message_type = message_json[0]
    if message_type == RelayMessageType.EVENT:
        event = Event.from_dict(message_json[2])
        if event.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE:
            if event.has_pubkey_ref(sender_pk.public_key.hex()):
                rdm = EncryptedDirectMessage.from_event(event)
                rdm.decrypt(sender_pk.hex(), public_key_hex=recipient.hex())
                print(f"New dm received:{event.date_time()} {rdm.cleartext_content}")
    elif message_type == RelayMessageType.OK:
        print(message_json)
    elif message_type == RelayMessageType.NOTICE:
        print(message_json)


if __name__ == "__main__":

    console = Console()

    pk = getpass.getpass(prompt="Enter sender private key (new one when empty): ")
    if len(pk) == 0:
        sender_pk = PrivateKey()
    else:
        sender_pk = PrivateKey.from_nsec(pk)

    print(f"New dm are sent from: {sender_pk.public_key.bech32()}")

    recipient_str = input("recipient (npub or nip05): ")
    recipient = ""
    if "@" in recipient_str:
        nip05 = extract_nip05(recipient_str)
        recipient = PublicKey.from_hex(nip05[0])
    elif "npub" in recipient_str:
        recipient = PublicKey.from_hex(recipient_str)
    if recipient != "":
        print(f"recipient is set to {recipient.bech32()}")
    else:
        raise Exception("reciever not valed")
    relay_url = input("relay: ")

    dm = EncryptedDirectMessage()
    dm.encrypt(
        sender_pk.hex(),
        cleartext_content="Hello, please reply to this dm for testing :).",
        recipient_pubkey=recipient.hex(),
    )

    filters = FiltersList(
        [
            Filters(
                authors=[recipient.hex()],
                kinds=[EventKind.ENCRYPTED_DIRECT_MESSAGE],
                since=get_timestamp(),
                limit=10,
            )
        ]
    )

    subscription_id = uuid.uuid1().hex
    io_loop = tornado.ioloop.IOLoop.current()
    message_pool = MessagePool(first_response_only=False)
    policy = RelayPolicy()
    r = Relay(
        relay_url,
        message_pool,
        io_loop,
        policy,
        timeout=5,
        close_on_eose=False,
        message_callback=print_dm,
    )
    dm_event = dm.to_event()
    dm_event.sign(sender_pk.hex())
    r.publish(dm_event.to_message())
    r.add_subscription(subscription_id, filters)

    try:
        io_loop.run_sync(r.connect)
    except gen.Return:
        pass
    io_loop.stop()



import logging

from tornado import gen
from tornado.ioloop import IOLoop
from tornado.websocket import WebSocketError, websocket_connect

from .base_relay import BaseRelay, RelayPolicy
from .message_pool import MessagePool

log = logging.getLogger(__name__)


class Relay(BaseRelay):
    def __init__(
        self,
        url: str,
        message_pool: MessagePool,
        io_loop: IOLoop,
        policy: RelayPolicy = RelayPolicy(),
        timeout: float = 2.0,
        close_on_eose: bool = True,
        message_callback=None,
    ) -> None:
        super().__init__(
            url,
            message_pool,
            policy,
            None,
            None,
            timeout,
            close_on_eose,
            message_callback,
        )
        self.io_loop = io_loop
        self.running = True

    @property
    def is_connected(self) -> bool:
        return self.ws is not None and self.ws.protocol is not None

    @gen.coroutine
    def connect(self):
        error = False
        try:
            if self.timeout > 0:
                self.ws = yield gen.with_timeout(
                    self.io_loop.time() + self.timeout,
                    websocket_connect(
                        self.url,
                        ping_interval=60,
                        ping_timeout=120,
                    ),
                )
            else:
                self.ws = yield websocket_connect(
                    self.url,
                    ping_interval=60,
                    ping_timeout=120,
                )
            self.connected = True
            # yield self.ws.write_message(self.request)
            self.publish(self.request)
            # self.io_loop.call_later(1, self.send_message, self.request)
            while True:
                if self.outgoing_messages.qsize() > 0:
                    message = self.outgoing_messages.get()
                    self.num_sent_events += 1
                    yield self.ws.write_message(message)
                message = yield self.ws.read_message()
                if message is None:
                    break
                self._on_message(message)
                if not self.connected:
                    break

        except gen.TimeoutError:
            log.warning(f"Timeout connecting to {self.url}")
            error = True
        except WebSocketError as e:
            log.warning(f"Error connecting to WebSocket server at {self.url}: {e}")
            error = True
        except Exception as e:
            log.warning(f"Error connecting to {self.url}: {e}")
            error = True
        if error:
            self.error_counter += 1
            if self.error_counter <= self.error_threshold:
                self.io_loop.call_later(1, self.connect)
            else:
                return
        # print(self.request)
        # self.publish(self.request)

        log.info(f"WebSocket connection to {self.url} closed")

    def _eose_received(self):
        if self.close_on_eose:
            self.close()

    @gen.coroutine
    def start(self):
        yield self.connect()

    @gen.coroutine
    def close(self):
        if self.ws is not None:
            self.connected = False
            self.error_counter = 0
            yield self.ws.close()
            # self.io_loop.stop()
